from __future__ import absolute_import, division, print_function

import time

# TODO: change thrown errors to these
from concurrent.futures import ALL_COMPLETED, CancelledError, TimeoutError

from ..messages import PythonTask
from ..scheduler import QueueScheduler, Running
from ..utils import timeout as seconds

__all__ = ('MesosPoolExecutor',
           'Future')


def wait(fs, timeout=None, return_when=ALL_COMPLETED):
    raise NotImplementedError()


def as_completed(fs, timeout=None):
    raise NotImplementedError()


class Future(object):

    def __init__(self, task):
        self.task = task

    @property
    def status(self):
        return self.task.status

    def cancel(self):
        raise NotImplementedError()

    def cancelled(self):
        return self.status.has_killed()

    def running(self):
        return (self.status.is_running() or
                self.status.is_starting() or
                self.status.is_staging())

    def done(self):
        return self.status.has_killed() or self.status.has_finished()

    def result(self, timeout=None):
        with seconds(timeout):
            while not self.status.has_terminated():
                time.sleep(0.1)
            if self.status.has_finished():
                return self.status.data
            else:
                raise self.status.data

    def exception(self, timeout=None):
        with seconds(timeout):
            while not self.status.has_terminated():
                time.sleep(0.1)
            if self.status.has_finished():
                return None
            else:
                return self.status.data

    def add_done_callback(self, fn):
        raise NotImplementedError()


class MesosPoolExecutor(Running):

    def __init__(self, max_workers=-1, *args, **kwargs):
        self.max_worker = max_workers  # TODO
        self.scheduler = QueueScheduler()
        super(MesosPoolExecutor, self).__init__(
            self.scheduler, *args, **kwargs)

    def submit(self, fn, args=[], kwargs={}, **kwds):
        task = PythonTask(fn=fn, args=args, kwargs=kwargs,
                          name=kwds.pop('name', 'futures'), **kwds)
        self.scheduler.submit(task)
        return Future(task)

    def map(self, func, *iterables, **kwargs):
        timeout = kwargs.pop('timeout', None)
        chunksize = kwargs.pop('chunksize', 1)
        for item in zip(*iterables):
            future = self.submit(func, args=item, **kwargs)
            yield future.result(timeout=timeout)

    def shutdown(self, wait=True):
        if wait:
            self.scheduler.wait(-1)
        self.stop()



from __future__ import absolute_import, division, print_function

import operator
import time
from collections import Iterator

import pytest
from satyr.apis.futures import Future, MesosPoolExecutor
from satyr.proxies.messages import Cpus, Disk, Mem
from satyr.utils import TimeoutError, timeout


@pytest.fixture
def resources():
    return [Cpus(0.1), Mem(128), Disk(0)]


def test_submit():
    with MesosPoolExecutor(name='futures-pool') as executor:
        future1 = executor.submit(operator.add, [1, 2])
        future2 = executor.submit(operator.mul, [3, 5])

        assert isinstance(future1, Future)
        assert isinstance(future2, Future)

        assert future1.result(timeout=5) == 3
        assert future2.result(timeout=5) == 15


def test_future_states():
    def add(a, b):
        time.sleep(2)
        return a + b

    with MesosPoolExecutor(name='futures-pool') as executor:
        future = executor.submit(add, [1, 2])
        with timeout(5):
            while future.running():
                time.sleep(0.1)
            assert future.running() is False
            assert future.done() is True
            assert future.result() == 3


def test_future_timeout():
    with pytest.raises(TimeoutError):
        with MesosPoolExecutor(name='futures-pool') as executor:
            future1 = executor.submit(time.sleep, [3])
            future1.result(timeout=1)


def test_future_raises_exception():
    def raiser():
        raise Exception('Boooom!')

    with MesosPoolExecutor(name='futures-pool') as executor:
        with pytest.raises(Exception) as e:
            future1 = executor.submit(raiser)
            future1.result(timeout=5)
            assert e.value.message == 'Boooom!'


def test_future_catches_exception():
    def raiser():
        raise Exception('Boooom!')

    with MesosPoolExecutor(name='futures-pool') as executor:
        future = executor.submit(raiser)
        e = future.exception(timeout=5)
        assert isinstance(e, Exception)
        assert e.message == 'Boooom!'


def test_multiple_submit(resources):
    def fn(a, b):
        return a + b

    with MesosPoolExecutor(name='futures-pool') as executor:
        futures = [executor.submit(fn, args=[1, i], resources=resources)
                   for i in range(10)]
        values = [f.result(timeout=5) for f in futures]

    assert values == [i + 1 for i in range(10)]


def test_map(resources):
    with MesosPoolExecutor(name='futures-pool') as executor:
        it = executor.map(operator.add, range(10), range(10),
                          resources=resources)
        assert isinstance(it, Iterator)
        for i, v in enumerate(it):
            assert i + i == v



from __future__ import absolute_import, division, print_function

import time
from Queue import Empty

import cloudpickle as cp
from kazoo.client import KazooClient
from kazoo.recipe.queue import LockingQueue as KazooLockingQueue
from kazoo.recipe.queue import Queue as KazooQueue

from .utils import timeout as seconds
from .utils import TimeoutError


class SerializableMixin(object):

    def __getstate__(self):
        hosts = ["{}:{}".format(h, p) for h, p in self.client.hosts]
        client = ",".join(hosts)

        result = self.__dict__.copy()
        result['client'] = client

        return result

    def __setstate__(self, state):
        hosts = state.pop('client')
        client = KazooClient(hosts)
        client.start()

        self.__dict__ = state
        self.client = client


class CompatMixin(object):  # Python's Queue compatibility

    def __bool__(self):
        return True

    def __nonzero__(self):
        return True

    def qsize(self):
        return len(self)

    def empty(self):
        return len(self) == 0


class Queue(CompatMixin, SerializableMixin, KazooQueue):

    def get(self, block=True, timeout=-1):
        result = super(Queue, self).get()

        if block:
            try:
                with seconds(timeout):
                    while result is None:
                        result = super(Queue, self).get()
                        time.sleep(0.1)
            except TimeoutError:
                raise Empty

        return cp.loads(result)

    def put(self, item):
        value = cp.dumps(item)
        return super(Queue, self).put(value)


class LockingQueue(CompatMixin, SerializableMixin, KazooLockingQueue):

    def get(self, block=True, timeout=-1):
        result = super(LockingQueue, self).get(timeout=timeout)
        return cp.loads(result)

    def put(self, item, priority=100):
        value = cp.dumps(item)
        return super(LockingQueue, self).put(value, priority=priority)



from __future__ import absolute_import, division, print_function

import time

from ..messages import PythonTask
from ..queue import Queue
from ..scheduler import QueueScheduler, Running
from ..utils import timeout

__all__ = ('Pool',
           'Queue',
           'AsyncResult')


class AsyncResult(object):

    def __init__(self, task):
        self.task = task

    @property
    def status(self):
        return self.task.status

    def get(self, timeout=60):
        self.wait(timeout)
        if self.successful():
            return self.status.data
        else:
            try:
                raise self.status.data
            except TypeError:
                raise ValueError('Async result indicate task failed!')

    def wait(self, seconds=60):
        with timeout(seconds):
            while not self.ready():
                time.sleep(0.1)

    def ready(self):
        return self.status.has_terminated()

    def successful(self):
        return self.status.has_succeeded()


class Pool(Running):

    def __init__(self, processes=-1, *args, **kwargs):
        self.processes = processes
        self.scheduler = QueueScheduler()
        super(Pool, self).__init__(self.scheduler, *args, **kwargs)

    def close(self):
        self.stop()

    def terminate(self):
        self.stop()

    def wait(self, seconds=-1):
        self.scheduler.wait(seconds)

    def map(self, func, iterable, chunksize=1, **kwargs):
        results = self.map_async(func, iterable, chunksize, **kwargs)
        return [result.get(timeout=-1) for result in results]

    def map_async(self, func, iterable, chunksize=1, callback=None, **kwargs):
        return [self.apply_async(func, (item,), **kwargs) for item in iterable]

    def apply(self, func, args=[], kwds={}, **kwargs):
        result = self.apply_async(func=func, args=args, kwds=kwds, **kwargs)
        return result.get(timeout=-1)

    def apply_async(self, func, args=[], kwds={}, callback=None, **kwargs):
        task = PythonTask(name=kwargs.pop('name', 'multiprocessing'),
                          fn=func, args=args, kwargs=kwds, **kwargs)
        self.scheduler.submit(task)
        return AsyncResult(task)



from __future__ import absolute_import, division, print_function

import atexit
import logging
import os
import signal
import time
from collections import Counter

from mesos.interface import mesos_pb2
from mesos.native import MesosSchedulerDriver

from .binpack import bfd
from .interface import Scheduler
from .proxies import SchedulerProxy
from .proxies.messages import FrameworkInfo, TaskInfo, encode
from .utils import timeout


class Running(object):

    def __init__(self, scheduler, name, user='', master=os.getenv('MESOS_MASTER'),
                 implicit_acknowledge=1, *args, **kwargs):
        scheduler = SchedulerProxy(scheduler)
        framework = FrameworkInfo(name=name, user=user, *args, **kwargs)
        self.driver = MesosSchedulerDriver(scheduler, encode(framework),
                                           master, implicit_acknowledge)

        def shutdown(signal, frame):
            self.stop()

        signal.signal(signal.SIGINT, shutdown)
        signal.signal(signal.SIGTERM, shutdown)
        atexit.register(self.stop)

    def run(self):
        return self.driver.run()

    def start(self):
        status = self.driver.start()
        assert status == mesos_pb2.DRIVER_RUNNING
        return status

    def stop(self):
        return self.driver.stop()

    def join(self):
        return self.driver.join()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()
        self.join()
        if exc_type:
            raise exc_type, exc_value, traceback


class QueueScheduler(Scheduler):

    # todo remove statuses or replace as a property wich plucks status from
    # tasks
    def __init__(self, *args, **kwargs):
        self.tasks = {}    # holding task_id => (task, status, result) pairs

    @property
    def statuses(self):
        return {task_id: task.status for task_id, task in self.tasks.items()}

    def is_idle(self):
        return not len(self.tasks)

    def report(self):
        states = [status.state for status in self.statuses.values()]
        counts = Counter(states)
        message = ', '.join(['{}: {}'.format(key, count)
                             for key, count in counts.items()])
        logging.info('Task states: {}'.format(message))

    def wait(self, seconds=-1):
        with timeout(seconds):
            while not self.is_idle():
                time.sleep(0.1)

    def submit(self, task):  # supports commandtask, pythontask etc.
        assert isinstance(task, TaskInfo)
        self.tasks[task.id] = task

    def on_offers(self, driver, offers):
        logging.info('Received offers: {}'.format(sum(offers)))
        self.report()

        # maybe limit to the first n tasks
        staging = [self.tasks[status.task_id]
                   for status in self.statuses.values() if status.is_staging()]
        # best-fit-decreasing binpacking
        bins, skip = bfd(staging, offers, cpus=1, mem=1)

        for offer, tasks in bins:
            try:
                for task in tasks:
                    task.slave_id = offer.slave_id
                    task.status.state = 'TASK_STARTING'
                # running with empty task list will decline the offer
                driver.launch(offer.id, tasks)
            except Exception:
                logging.exception('Exception occured during task launch!')

    def on_update(self, driver, status):
        task = self.tasks[status.task_id]
        logging.info('Updated task {} state {}'.format(status.task_id,
                                                       status.state))
        if status.has_terminated():
            del self.tasks[task.id]
            if status.has_failed():
                logging.error('Aborting because task {} is in unexpected state '
                              '{} with message {}'.format(status.task_id,
                                                          status.state,
                                                          status.message))
                driver.abort()

        task.update(status)
        self.report()


if __name__ == '__main__':
    scheduler = QueueScheduler()
    with Running(scheduler, name='test') as fw:
        scheduler.wait()



from __future__ import absolute_import, division, print_function

import logging

import cloudpickle
from mesos.interface import mesos_pb2

from .proxies.messages import (CommandInfo, ContainerInfo, Cpus, Disk,
                               DockerInfo, Environment, ExecutorInfo, Mem,
                               TaskInfo, TaskStatus)


class PickleMixin(object):

    @property
    def data(self):
        return cloudpickle.loads(self['data'])

    @data.setter
    def data(self, value):
        self['data'] = cloudpickle.dumps(value)


class PythonTaskStatus(PickleMixin, TaskStatus):

    proto = mesos_pb2.TaskStatus(
        labels=mesos_pb2.Labels(
            labels=[mesos_pb2.Label(key='python')]))

    def __init__(self, data=None, **kwargs):
        super(PythonTaskStatus, self).__init__(**kwargs)
        self.data = data


class PythonTask(PickleMixin, TaskInfo):

    proto = mesos_pb2.TaskInfo(
        labels=mesos_pb2.Labels(
            labels=[mesos_pb2.Label(key='python')]))

    def __init__(self, fn=None, args=[], kwargs={},
                 resources=[Cpus(0.1), Mem(128), Disk(0)],
                 command='python -m satyr.executor', envs={}, uris=[],
                 docker='lensa/satyr:latest', **kwds):
        super(PythonTask, self).__init__(**kwds)
        self.status = PythonTaskStatus(task_id=self.id, state='TASK_STAGING')
        self.executor = ExecutorInfo(
            container=ContainerInfo(type='DOCKER', docker=DockerInfo(
                force_pull_image=False, network='HOST')),
            command=CommandInfo(shell=True))
        self.data = (fn, args, kwargs)
        self.envs = envs
        self.uris = uris
        self.docker = docker
        self.command = command
        self.resources = resources

    @property
    def uris(self):
        return [uri.value for uri in self.executor.command.uris]

    @uris.setter
    def uris(self, value):
        self.executor.command.uris = [{'value': v} for v in value]

    @property
    def envs(self):
        envs = self.executor.command.environment.variables
        return {env.name: env.value for env in envs}

    @envs.setter
    def envs(self, value):
        envs = [{'name': k, 'value': v} for k, v in value.items()]
        self.executor.command.environment = Environment(variables=envs)

    @property
    def command(self):
        return self.executor.command.value

    @command.setter
    def command(self, value):
        self.executor.command.value = value

    @property
    def docker(self):
        return self.executor.container.docker.image

    @docker.setter
    def docker(self, value):
        self.executor.container.docker.image = value

    def __call__(self):
        fn, args, kwargs = self.data
        return fn(*args, **kwargs)

    def update(self, status):
        assert isinstance(status, PythonTaskStatus)
        self.on_update(status)
        if status.has_succeeded():
            self.on_success(status)
        elif status.has_failed():
            self.on_fail(status)

    def on_update(self, status):
        self.status = status
        logging.info('Task {} has been updated with state {}'.format(
            self.id.value, status.state))

    def on_success(self, status):
        logging.info('Task {} has been succeded'.format(self.id.value))

    def on_fail(self, status):
        logging.info('Task {} has been failed due to {}'.format(
            self.id.value, status.message))
